( 3iPort.fs - words to read/write to ports in binary )
\ Assad Ebrahim
\ 2020-09-09---Wed---01:31 - started testing the ICs in Forth
\ 2020-09-12---Sat---23:34 - finished 

\ FEATURE: Built 4-port OUT using 74154 IC (4-bits to 16-line decoder)
\ FEATURE: Built 4-port IN using 2x cascaded 74148 ICs and a 7408 (AND) IC into a 16-line to 4-bits priority encoder

\ Code to test writing coded 4-bit to port and using 74154 IC to decode and turn on LED with associated address.  Saves I/O pins
\ on Nano controller (4 pins for 16 lines instead of 16 I/O lines).

\ Hardware: 
\ Arduino Nano with 3iForth.avr
\ connected to 74154 IC, with 2345 on Nano connecting to ABCD on chip (D = msb), 6 on Nano -> G2 control signal on chip
\ G1 control signal on chip pulled low
\ LEDs connected via 860R resistor from VCC (5V) to output addresses of chip (0-12)

\ Usage: 
\ n >>p  works to send the number (0-12) in binary to 74154 chip which decodes to pull the right line hi/lo and light led
\

\ Physical Port Control
: port4-on  \ for 74154 chip from Nano: 2345 are binary pins 5=msb, 6 is decode control signal - pull low to start.  
\       PORT IS ACTIVE HIGH
    2 3 4 5 6 out out out out out ;  
: >p 20 ms  6 lo ;  \ decode-start  all commands are now real-time reflected by the IC
: p> 20 ms  6 hi ;  \ decode stop  all commands are now reflected by the IC

\ Virtual Port Memory
create port4 4 bytes allot
variable pport  \ pointer to port
: reset-pport  port4 pport ! ;
reset-pport
: port4. reset-pport pport @ 4 0 do c@+ . loop drop ;  \ print virtual port contents

\ Extracting Bitfield of a 4-bit number into Virtual Port memory
:  T   $F  ;                        \ 1111b  Determines whether calculations are 4-, 8-, or 16-bits
:  ~    ( p -- ~p ) T xor ;         \ negation op takes bitfield size from T
: init-mask  T 2 / ~  ;  \ initial mask starts with MSB i.e. is $8=1000b, $80=1000000b, or $8000 depending on T
: .bit  ( n -- ) 0> if 1 else 0 then pport @ c!+ pport ! ;
: bits  (  bitfield  --  )  
    reset-pport init-mask begin      \  loop through bits, picking them off
        2dup and .bit    
        1 rshift dup 0= until 
    2drop 
    port4. ." b" 
    ;

\ Sending virtual port to physical
: bit>p ( 1/0 -- ) if hi else lo then ;
: >>p ( n -- ) bits p> reset-pport pport @ 4 0 do c@+ 5 i - swap bit>p loop drop >p ;

\ Test routines
: port-test   \ manual validation
    p>   all-low                 >p
    p>   2 hi                    >p
    p>   2 lo 3 hi               >p
    p>   2 hi                    >p
    p>   2 lo 3 lo 4 hi          >p
    p>   2 hi                    >p
    p>   2 lo 3 hi               >p
    p>   2 hi                    >p
    p>   2 lo 3 lo 4 lo 5 hi     >p
    p>   2 hi                    >p
    p>   2 lo 3 hi               >p
    p>   2 hi                    >p
    p>   2 lo 3 lo 4 hi          >p
    p>
; 

: port-test2 13 0 do i >>p loop p> ;  \ auto validation

: key-test  \ piano application validation
    \ white keys
    0 >>p 
    2 >>p
    4 >>p
    5 >>p
    7 >>p
    9 >>p
    11 >>p
    12 >>p
    \ black keys
    1 >>p
    3 >>p
    6 >>p
    8 >>p
    10 >>p
    p>
;

\ Hardware: 
\ Arduino Nano with 3iForth.avr
\ connected to 2x 74148 ICs cascaded with 7408 (AND), with pins 9,10,11,12 as data pins (lsb-msb) and pin 8 as enable (pull low)

\ Usage: 
\ port4 port4-in   setup ports on Arduino
\ port4read-test   continuously reads the 4-bit input port.  Test by setting 0-7 active low on 74148#1 then 8-15 low on 74148#2
\ Ctrl-C to exit the infinite loop

: alow ( 0/1 -- 1/0 ) 1 + 2 mod ;  \ active low
\ for each individual 74148 IC (8 line to 3 bit priority encoder)
: port3A 10 11 12 9 ;
: port3B 15 16 17 14 ;
: port3-in ( port -- ) in in in in ;
: 3drop 2drop drop ;
: read3 ( port -- lsb0 1 msb2 ) readpin alow if rot readpin alow -rot swap readpin alow swap readpin alow else 3drop then ;
: read3A port3A read3 ;
: read3B port3B read3 ;
: 3.  . . . ; 

\ for cascaded 2x 74148 ICs into 7408 AND chip (16 line to 4 bit priority encoder)
: port4 ( -- lsb ... msb valid )8 9 10 11  12 ;  \ pin12 = data valid pin (valid if active low), pins 8-11 are 4-bit port read
: port4-in ( port -- ) port4 in in in in in ;
: 4drop 2drop 2drop ;
: [read-port4] ( msb ... lsb -- )  4 0 do readpin alow . loop ;
: read4 ( port -- ) readpin alow if  [read-port4] else 4drop then ;

: port4-read  cr begin port4 readpin alow if [read-port4] cr else 4 drop then 0 until ;

: help-port
   cr cr
   cr ." PORT HELP"
   cr ." Dependencies: 3iForth.avr, 3iForth.fs, 3iLight.fs, 3iPort.fs"
   cr 
   cr ." 4-bit port OUT using 74154 IC (4-bit to 16 line decoder)"
   cr ." Active high inputs, active low outputs."
   cr ." port4-on - set pins 2,3,4,5,6 as 4-port OUT and control pin"
   cr ." n >>p  - send number to port"
   cr ." p>   - port off"
   cr ." >p   - port on"
   cr
   cr ." 4-bit port IN using 2x 74148 IC cascaded into 7408 AND chip (16-line to 4-bit priority encoder)"
   cr ." Active low inputs and active low outputs."
   cr ." port4-in - set pins 8,9,10,11,12 as 4-port IN and data valid pin"
   cr ." port4read-test - reads port printing encoded inputs until Ctrl-C pressed"
   cr
;

help-port
